# -*- coding: utf-8 -*-

# Script inspired from https://github.com/jts/nanopolish/blob/master/scripts/calculate_methylation_frequency.py

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
# Standard library imports
from collections import *
import csv
import datetime

# Third party imports
from tqdm import tqdm
import numpy as np

# Local imports
from NanopolishComp.common import *
from NanopolishComp import __version__ as package_version
from NanopolishComp import __name__ as package_name

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~MAIN CLASS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#

class Freq_meth_calculate():

    def __init__ (self,
        input_fn:"str",
        output_bed_fn:"str"="",
        output_tsv_fn:"str"="",
        min_depth:"int"=10,
        sample_id:"str"="",
        verbose:"bool"=False,
        quiet:"bool"=False):
        """
        Calculate methylation frequency at genomic CpG sites from the output of nanopolish call-methylation
        * input_fn
            Path to a nanopolish call_methylation tsv output file
        * output_bed_fn
            Path to write a summary result file in BED format
        * output_tsv_fn
            Path to write an more extensive result report in TSV format
        * min_depth
            Minimal number of reads covering a site to be reported
        * sample_id
            Sample ID to be used for the bed track header
        * verbose
            Increase verbosity
        * quiet
            Reduce verbosity
        """

        # Save init options in dict for later
        kwargs = locals()

        # Define overall verbose level
        self.log = get_logger(name="Freq_meth_calculate", verbose=verbose, quiet=quiet)

        # Print option summary log
        self.log.debug ("## Options summary ##")
        self.log.debug ("\tpackage_name: {}".format(package_name))
        self.log.debug ("\tpackage_version: {}".format(package_version))
        self.log.debug ("\ttimestamp: {}".format(str(datetime.datetime.now())))
        self.log.debug (dict_to_str(kwargs, nsep=1, exclude_list=["self"]))

        # Verify parameters validity
        self.log.warning ("## Checking arguments ##")

        # Try to read input file if not a stream
        self.log.debug("\tTesting input file readability")
        if input_fn != 0 and not file_readable (input_fn):
            raise IOError ("Cannot read input file")

        # Verify that at least one output file is given:
        self.log.debug("\tCheck output file")
        if not output_bed_fn and not output_tsv_fn:
            raise NanopolishCompError("At least one output file should be given")
        if output_bed_fn:
            if os.path.dirname(output_bed_fn):
                mkdir (os.path.dirname(output_bed_fn), exist_ok=True)
            self.log.debug("\t\tOutput results in bed format")
        if output_tsv_fn:
            if os.path.dirname(output_tsv_fn):
                mkdir (os.path.dirname(output_tsv_fn), exist_ok=True)
            self.log.debug("\t\tOutput results in tsv format")

        # Create self variables
        self.counter = Counter()
        self.input_fn = input_fn
        self.output_bed_fn = output_bed_fn
        self.output_tsv_fn = output_tsv_fn
        self.min_depth = min_depth
        self.sample_id = sample_id

        self.log.warning ("## Parsing methylation_calls file ##")
        self._parse_methylation_calls ()
        self.log.info ("## Results summary ##")
        self.log.info (dict_to_str(self.counter, nsep=1))

    def _parse_methylation_calls(self):
        """"""

        # Create collection to store results
        site_dict = defaultdict(list)

        try:
            input_fp = open (self.input_fn, "r")
            self.log.debug ("\tWrite output file header")
            if self.output_bed_fn:
                output_bed_fp = open (self.output_bed_fn, "w")
                output_bed_fp.write(Site.BED_header(self.sample_id)+"\n")
            if self.output_tsv_fn:
                output_tsv_fp = open (self.output_tsv_fn, "w")
                output_tsv_fp.write(Site.TSV_header()+"\n")

            self.log.info ("\tStarting to parse file Nanopolish methylation call file")
            header_line = input_fp.readline()
            byte_offset = len(header_line)
            lp = LineParser(header_line, sep="\t", cast_numeric_field=True)

            for line in tqdm(input_fp, desc="\t", unit=" lines", disable=self.log.level>=30):
                self.counter["Total read lines"]+=1
                byte_len = len(line)
                l = lp(line)

                if not l:
                    # Failsafe if line is malformed
                    self.counter["Invalid read line"]+=1
                else:
                    # Store byte offset corresponding to appropriate line
                    self.counter["Valid read lines"]+=1
                    site_dict[(l.chromosome, l.start, l.strand)].append(byte_offset)
                    byte_offset += byte_len

            self.log.info ("\tFiltering out low coverage sites")
            filtered_site_dict = defaultdict(list)
            for k, offset_list in site_dict.items():
                self.counter["Total sites"]+=1

                # If low coverage unset list to release memory
                if len(offset_list) < self.min_depth:
                    self.counter["Low coverage sites"]+=1
                else:
                    self.counter["Valid sites"]+=1
                    filtered_site_dict[k]=offset_list
            del site_dict

            self.log.info ("\tSorting by coordinates")
            filtered_site_dict = OrderedDict(sorted(filtered_site_dict.items(), key=lambda t: t[0]))

            self.log.info ("\tProcessing valid sites found")
            for k, offset_list in tqdm(filtered_site_dict.items(), desc="\t", unit=" sites", disable=self.log.level>=30):
                # Get all read lines corresponding to current site
                ll = []
                for offset in offset_list:
                    input_fp.seek(offset, 0)
                    ll.append (lp(input_fp.readline()))

                # Parse list with helper class Site
                site = Site (ll)
                self.counter["Valid sites"]+=1
                if self.output_bed_fn:
                    output_bed_fp.write(site.to_bed()+"\n")
                if self.output_tsv_fn:
                    output_tsv_fp.write(site.to_tsv()+"\n")
        finally:
            input_fp.close()
            if self.output_bed_fn:
                output_bed_fp.close()
            if self.output_tsv_fn:
                output_tsv_fp.close()

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~HELPER CLASS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#

class Site ():
    """Structure like class to store site information"""

    # Class id dispatcher
    ID = 0
    @classmethod
    def next_id (cls):
        id = cls.ID
        cls.ID +=1
        return id

    @classmethod
    def BED_header (cls, sample_id=""):
        return "track name=Nanopolish_methylation_{} itemRgb=On".format(sample_id)

    @classmethod
    def TSV_header (cls):
        return "\t".join(["chromosome","start","end","strand","site_id","methylated_reads","unmethylated_reads","ambiguous_reads","sequence","num_motifs","llr_list"])

    def __init__ (self, ll, min_llr=2):
        """"""
        self.total = len(ll)
        self.methylated = 0
        self.unmethylated = 0
        self.ambiguous = 0
        self.id = self.next_id()
        self.sequence = ll[0].sequence
        self.num_motifs = ll[0].num_motifs
        self.chromosome = ll[0].chromosome
        self.start = ll[0].start
        self.end = ll[0].end+1
        self.strand = ll[0].strand

        llr_list = []
        for l in ll:
            llr_list.append(float(l.log_lik_ratio))
            # Count read methylation call per site
            if l.log_lik_ratio >= min_llr:
                self.methylated+=1
            elif l.log_lik_ratio <= -min_llr:
                self.unmethylated+=1
            else:
                self.ambiguous+=1

        self.med_llr = np.mean(llr_list)
        self.llr_list = llr_list
        if self.med_llr <= -min_llr:
            self.color = '8,121,207'
        elif self.med_llr < min_llr:
            self.color = '100,100,100'
        else:
            self.color = '235,5,79'

    def __repr__(self):
        return "{}:{}-{}({}) / id:{} / reads:{}".format(
            self.chromosome,
            self.start,
            self.end,
            self.strand,
            self.id,
            self.total)

    def to_bed (self):
        """"""
        return "{}\t{}\t{}\t{}\t{:.6f}\t{}\t{}\t{}\t'{}''".format(
            self.chromosome,
            self.start,
            self.end,
            self.id,
            self.med_llr,
            self.strand,
            self.start,
            self.end,
            self.color)

    def to_tsv (self):
        """"""
        return "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}".format(
            self.chromosome,
            self.start,
            self.end,
            self.strand,
            self.id,
            self.methylated,
            self.unmethylated,
            self.ambiguous,
            self.sequence,
            self.num_motifs,
            ",".join([str(i) for i in self.llr_list]))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~#
# Standard library imports
import argparse
import sys

# Local imports
from NanopolishComp import __version__ as package_version
from NanopolishComp import __name__ as package_name
from NanopolishComp import __description__ as package_description
from NanopolishComp.Eventalign_collapse import Eventalign_collapse
from NanopolishComp.Freq_meth_calculate import Freq_meth_calculate

#~~~~~~~~~~~~~~TOP LEVEL ENTRY POINT~~~~~~~~~~~~~~#
def main(args=None):
    """
    Main entry point for NanoSnake command line interface
    """

    # Parser and subparsers for command
    parser = argparse.ArgumentParser (description=package_description)
    parser.add_argument("--version", action="version", version="{} v{}".format(package_name, package_version))
    subparsers = parser.add_subparsers (description="%(prog)s implements the following subcommands", dest="subcommands")
    subparsers.required = True

    # Eventalign_collapse subparser
    subparser_ec = subparsers.add_parser("Eventalign_collapse", description="Collapse the nanopolish eventalign output at kmers level and compute kmer level statistics")
    subparser_ec.set_defaults(func=Eventalign_collapse_main)
    subparser_ec_io = subparser_ec.add_argument_group("Input/Output options")
    subparser_ec_io.add_argument("-i", "--input_fn", default=0, help="Path to a nanopolish eventalign tsv output file. If '0' read from std input (default: %(default)s)")
    subparser_ec_io.add_argument("-o", "--outdir", type=str, default="./", help="Path to the output folder (will be created if it does exist yet) (default: %(default)s)")
    subparser_ec_io.add_argument("-p", "--outprefix", type=str, default="out", help="text outprefix for all the files generated (default: %(default)s)")
    subparser_ec_rp = subparser_ec.add_argument_group("Run parameters options")
    subparser_ec_rp.add_argument("-s", "--write_samples", default=False, action='store_true', help="If given, will write the raw sample if nanopolish eventalign was ran with --samples option (default: %(default)s)")
    subparser_ec_rp.add_argument("-r", "--max_reads", default=0 , type=int , help = "Maximum number of read to parse. 0 to deactivate (default: %(default)s)")
    subparser_ec_rp.add_argument("-f", "--stat_fields", default=["mean", "median", "num_signals"], type=str, nargs='+', help = "List of statistical fields to compute if nanopolish eventalign was ran with --sample option. Valid values = mean, std, median, mad, num_signals (default: %(default)s)")
    subparser_ec_other = subparser_ec.add_argument_group("Other options")
    subparser_ec_other.add_argument("-t", "--threads", default=4, type=int, help="Total number of threads. 1 thread is used for the reader and 1 for the writer (default: %(default)s)")

    # Freq_meth_calculate subparser
    subparser_fm = subparsers.add_parser("Freq_meth_calculate", description="Calculate methylation frequency at genomic CpG sites from the output of nanopolish call-methylation")
    subparser_fm.set_defaults(func=Freq_meth_calculate_main)
    subparser_fm_io = subparser_fm.add_argument_group("Input/Output options")
    subparser_fm_io.add_argument("-i", "--input_fn", default=0, help="Path to a nanopolish call_methylation tsv output file. If not specified read from std input")
    subparser_fm_io.add_argument("-b", "--output_bed_fn", type=str, default="", help="Path to write a summary result file in BED format (default: %(default)s)")
    subparser_fm_io.add_argument("-t", "--output_tsv_fn", type=str, default="", help="Path to write an more extensive result report in TSV format (default: %(default)s)")
    subparser_fm_fo = subparser_fm.add_argument_group("Filtering options")
    subparser_fm_fo.add_argument("-d", "--min_depth", type=int, default=10, help="Minimal number of reads covering a site to be reported (default: %(default)s)")
    subparser_fm_other = subparser_fm.add_argument_group("Other options")
    subparser_fm_other.add_argument("-s", "--sample_id", type=str, default="", help="Sample ID to be used for the bed track header (default: %(default)s)")

    # Add common group parsers
    for sp in [subparser_ec, subparser_fm]:
        sp_verbosity = sp.add_mutually_exclusive_group()
        sp_verbosity.add_argument("-v", "--verbose", action="store_true", default=False, help="Increase verbosity (default: %(default)s)")
        sp_verbosity.add_argument("-q", "--quiet", action="store_true", default=False, help="Reduce verbosity (default: %(default)s)")

    # Parse args and call subfunction
    args = parser.parse_args()
    args.func(args)

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~SUBPARSERS FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~~~~~~~#

def Eventalign_collapse_main (args):
    """"""
    # Run corresponding class
    Eventalign_collapse (
        input_fn = args.input_fn,
        outdir = args.outdir,
        outprefix = args.outprefix,
        write_samples = args.write_samples,
        stat_fields= args.stat_fields,
        threads = args.threads,
        verbose = args.verbose,
        quiet = args.quiet)

def Freq_meth_calculate_main (args):
    """"""
    # Run corresponding class
    Freq_meth_calculate (
        input_fn = args.input_fn,
        output_bed_fn = args.output_bed_fn,
        output_tsv_fn = args.output_tsv_fn,
        min_depth = args.min_depth,
        sample_id = args.sample_id,
        verbose = args.verbose,
        quiet = args.quiet)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup

# Define package info
name = "NanopolishComp"
version = "0.6.9"
description = 'NanopolishComp is a Python3 package for downstream analyses of Nanopolish output files'
with open("README.md", "r") as fh:
    long_description = fh.read()

# Collect info in a dictionnary for setup.py
setup(
    name = name,
    description = description,
    version = version,
    long_description = long_description,
    long_description_content_type="text/markdown",
    url = "https://github.com/a-slide/NanopolishComp",
    author = 'Adrien Leger',
    author_email = 'aleg@ebi.ac.uk',
    license= "MIT",
    python_requires ='>=3.6',
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'],
    install_requires = ['numpy>=1.14.0', 'tqdm>=4.23.4'],
    packages = [name],
    entry_points = {'console_scripts': ['NanopolishComp=NanopolishComp.__main__:main']})



# -*- coding: utf-8 -*-

#~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~#
# Standard library imports
import sys
import os
from collections import *
import logging

#~~~~~~~~~~~~~~FUNCTIONS~~~~~~~~~~~~~~#

def stderr_print (*args):
    """reproduce print with stderr.write"""
    sys.stderr.write(" ".join(str(a) for a in args))
    sys.stderr.flush()

def file_readable (fn, **kwargs):
    """Check if the file is readable"""
    return os.path.isfile (fn) and os.access (fn, os.R_OK)

def dir_writable (fn, **kwargs):
    """Check if the file is readable"""
    if not os.path.isdir(fn):
        fn = os.path.dirname(fn)
    return os.path.dirname(fn) and os.access (fn, os.W_OK)

def mkdir (fn, exist_ok=False):
    """ Create directory recursivelly. Raise IO error if path exist or if error at creation """
    try:
        os.makedirs (fn, exist_ok=exist_ok)
    except:
        raise NanopolishCompError ("Error creating output folder `{}`".format(fn))

def numeric_cast_dict (d):
    """Cast str values to integer or float from a dict """
    for k, v in d.items():
        d[k] = numeric_cast(v)
    return d

def numeric_cast (v):
    """Try to cast values to int or to float"""
    if type(v)== str:
        try:
            v = int(v)
        except ValueError:
            try:
                v = float(v)
            except ValueError:
                pass
    return v

def dict_to_str (d, sep="\t", nsep=0, exclude_list=[]):
    """ Transform a multilevel dict to a tabulated str """
    m = ""

    if isinstance(d, Counter):
        for i, j in d.most_common():
            if not i in exclude_list:
                m += "{}{}: {:,}\n".format(sep*nsep, i, j)

    else:
        for i, j in d.items():
            if not i in exclude_list:
                if isinstance(j, dict):
                    j = dict_to_str(j, sep=sep, nsep=ntab+1)
                    m += "{}{}\n{}".format(sep*nsep, i, j)
                else:
                    m += "{}{}: {}\n".format(sep*nsep, i, j)
    return m[:-1]

def jhelp (f:"python function or method"):
    """
    Display a Markdown pretty help message for functions and class methods (default __init__ is a class is passed)
    jhelp also display default values and type annotations if available.
    Undocumented options are not displayed.
    The docstring synthax should follow the markdown formated convention below
    * f
        Function or method to display the help message for
    """
    # For some reason signature is not always importable. In these cases the build-in help is called instead
    try:
        from IPython.core.display import display, Markdown, HTML
        import inspect
    except (NameError, ImportError) as E:
        NanocomporeWarning ("jupyter notebook is required to use this function. Please verify your dependencies")
        help(f)
        return

    if inspect.isclass(f):
        f = f.__init__

    if inspect.isfunction(f) or inspect.ismethod(f):

        # Parse arguments default values and annotations
        sig_dict = OrderedDict()
        for name, p in inspect.signature(f).parameters.items():
            sig_dict[p.name] = []
            # Get Annotation
            if p.annotation != inspect._empty:
                sig_dict[p.name].append(": {}".format(p.annotation))
            # Get default value if available
            if p.default == inspect._empty:
                sig_dict[p.name].append("(required)")
            else:
                sig_dict[p.name].append("(default = {})".format(p.default))

        # Parse the docstring
        doc_dict = OrderedDict()
        descr = []
        lab=None
        for l in inspect.getdoc(f).split("\n"):
            l = l.strip()
            if l:
                if l.startswith("*"):
                    lab = l[1:].strip()
                    doc_dict[lab] = []
                elif lab:
                    doc_dict[lab].append(l)
                else:
                    descr.append(l)

        # Reformat collected information in Markdown synthax
        s = "---\n\n**{}.{}**\n\n{}\n\n---\n\n".format(f.__module__, f.__name__, " ".join(descr))
        for k, v in doc_dict.items():
            s+="* **{}** *{}*\n\n{}\n\n".format(k, " ".join(sig_dict[k]), " ".join(v))

        # Display in Jupyter
        display (Markdown(s))

def head (fp, n=10, sep="\t", comment=None):
    """
    Emulate linux head cmd. Handle gziped files and bam files
    * fp
        Path to the file to be parse.
    * n
        Number of lines to print starting from the begining of the file (Default 10)
    """
    line_list = []

    # Get lines
    try:
        with open(fp) as fh:
            line_num = 0
            while (line_num < n):
                l= next(fh).strip()
                if comment and l.startswith(comment):
                    continue
                if sep:
                    line_list.append (l.split(sep))
                else:
                    line_list.append (l)
                line_num+=1

    except StopIteration:
        pass

    # Add padding if sep given
    if sep:
        try:
            # Find longest elem per col
            col_len_list = [0 for _ in range (len(line_list[0]))]
            for ls in line_list:
                for i in range (len(ls)):
                    len_col = len(ls[i])
                    if len_col > col_len_list[i]:
                        col_len_list[i] = len_col

            # Add padding
            line_list_tab = []
            for ls in line_list:
                s = ""
                for i in range (len(ls)):
                    len_col = col_len_list[i]
                    len_cur_col = len(ls[i])
                    s += ls[i][0:len_col] + " "*(len_col-len_cur_col)+" "
                line_list_tab.append(s)
            line_list = line_list_tab

        # Fall back to non tabulated display
        except IndexError:
            return head (fp=fp, n=n, sep=None)

    for l in line_list:
        print (l)
    print()

def get_logger (name=None, verbose=False, quiet=False):
    """Set logger to appropriate log level"""

    logging.basicConfig(format='%(message)s')
    logger = logging.getLogger(name)

    # Define overall verbose level
    if verbose:
        logger.setLevel(logging.DEBUG)
    elif quiet:
        logger.setLevel(logging.WARNING)
    else:
        logger.setLevel(logging.INFO)

    return logger

class LineParser():
    """Simple line parser which returns namedtuples per line and does numeric
    type casting for file lines"""

    def __init__(self, header_line, sep="\t", cast_numeric_field=True):
        """"""
        self.cast_numeric_field = cast_numeric_field
        self.sep = sep
        self.header = header_line.strip().split(self.sep)
        self.line_tuple = namedtuple("Line", self.header)
        self.c = Counter()

    def __repr__ (self):
        m = "Header names: {}\n".format(self.header)
        m+="Line counter:{}".format(dict_to_str(self.c))
        return m

    def __call__(self, line):
        """"""
        line = line.strip().split(self.sep)

        # Try to cast numeric field to appropiate type
        if self.cast_numeric_field:
            for i in range(len(line)):
                line[i] = self._numeric_cast(line[i])

        if len(line) != len(self.header):
            self.c["Inconsistent field numbers"]+=1
            return None
        else:
            # Cast line_tuple object
            line = self.line_tuple(*line)
            self.c["Valid line parsed"]+=1
            return line

    def _numeric_cast(self, val):
        """Try to cast values to int or to float"""
        if type(val)== str:
            try:
                val = int(val)
            except ValueError:
                try:
                    val = float(val)
                except ValueError:
                    pass
        return val


#~~~~~~~~~~~~~~CUSTOM EXCEPTION AND WARN CLASSES~~~~~~~~~~~~~~#
class NanopolishCompError (Exception):
    """ Basic exception class for NanopolishComp package """
    pass



# -*- coding: utf-8 -*-

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#

# Disable multithreading for MKL and openBlas
import os
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["MKL_THREADING_LAYER"] = "sequential"
os.environ["NUMEXPR_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"
os.environ['OPENBLAS_NUM_THREADS'] = '1'

# Standard library imports
import multiprocessing as mp
from time import time
from collections import *
import traceback
import datetime

# Third party imports
import numpy as np
from tqdm import tqdm

# Local imports
from NanopolishComp.common import *
from NanopolishComp import __version__ as package_version
from NanopolishComp import __name__ as package_name

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~LOGGING INFO~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
import logging
logging.basicConfig(level=logging.INFO, format="%(message)s")

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~MAIN CLASS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#

class Eventalign_collapse ():

    def __init__ (self,
        input_fn:"str",
        outdir:"str"="./",
        outprefix:"str"="out",
        max_reads:"int"=None,
        write_samples:"bool"=False,
        stat_fields:"list of str"=["mean", "median", "num_signals"],
        threads:"int"=4,
        verbose:"bool"=False,
        quiet:"bool"=False):
        """
        Collapse the nanopolish eventalign output by kmers rather that by events.
        kmer level statistics (mean, median, std, mad) are only computed if nanopolish is run with --samples option
        * input_fn
            Path to a nanopolish eventalign tsv output file.
        * outdir
            Path to the output folder (will be created if it does exist yet)
        * outprefix
            text outprefix for all the files generated
        * max_reads
            Maximum number of read to parse. 0 to deactivate (default = 0)
        * write_samples
            If given, will write the raw sample if nanopolish eventalign was ran with --samples option
        * stat_fields
            List of statistical fields to compute if nanopolish eventalign was ran with --samples option.
            Valid values = "mean", "std", "median", "mad", "num_signals"
        * threads
            Total number of threads. 1 thread is used for the reader and 1 for the writer (default = 4)
        * verbose
            Increase verbosity
        * quiet
            Reduce verbosity
        """

        # Save init options in dict for later
        kwargs = locals()

        # Define overall verbose level
        self.log = logging.getLogger()
        if verbose:
            self.log.setLevel (logging.DEBUG)
        elif quiet:
            self.log.setLevel (logging.WARNING)
        else:
            self.log.setLevel (logging.INFO)

        # Collect args in dict for log report
        self.option_d = OrderedDict()
        self.option_d["package_name"] = package_name
        self.option_d["package_version"] = package_version
        self.option_d["timestamp"] = str(datetime.datetime.now())
        for i, j in kwargs.items():
            if i != "self":
                self.option_d[i]=j
        self.log.debug ("Options summary")
        self.log.debug (dict_to_str(self.option_d))

        # Verify parameters validity
        self.log.info ("Checking arguments")
        # Try to read input file if not a stream
        self.log.debug("\tTesting input file readability")
        if input_fn != 0 and not file_readable (input_fn):
            raise IOError ("Cannot read input file")
        # Try to create output folder
        self.log.debug("\tCreating output folder")
        mkdir(outdir, exist_ok=True)
        # Check other args
        self.log.debug("\tChecking number of threads")
        if threads < 3:
            raise ValueError ("At least 3 threads required")
        self.log.debug("\tChecking if stat_fields names are valid")
        for field in stat_fields:
            if not field in ["mean", "std", "median", "mad", "num_signals"]:
                raise ValueError ("Invalid value in stat_field {}. Valid entries = mean, std, median, mad, num_signals".format(field))

        # Save args to self values
        self.outdir = outdir
        self.outprefix = outprefix
        self.input_fn = input_fn
        self.threads = threads-2 # Remove 2 threads for read and write
        self.max_reads = max_reads
        self.write_samples = write_samples
        self.stat_fields = stat_fields

        # Init Multiprocessing variables
        in_q = mp.Queue (maxsize = 1000)
        out_q = mp.Queue (maxsize = 1000)
        error_q = mp.Queue ()

        # Define processes
        ps_list = []
        ps_list.append (mp.Process (target=self._split_reads, args=(in_q, error_q)))
        for i in range (self.threads):
            ps_list.append (mp.Process (target=self._process_read, args=(in_q, out_q, error_q, i+1)))
        ps_list.append (mp.Process (target=self._write_output, args=(out_q, error_q)))

        self.log.info ("Starting to process files")
        try:
            # Start all processes
            for ps in ps_list:
                ps.start ()
            # Monitor error queue
            for E in iter (error_q.get, None):
                raise E
            # Join processes
            for ps in ps_list:
                ps.join ()

        # Kill processes if any error
        except (BrokenPipeError, KeyboardInterrupt, NanopolishCompError) as E:
            for ps in ps_list:
                ps.terminate ()
            self.log.warning ("\nAn error occured. All processes were killed\n")
            raise E

    def __repr__ (self):
        m = "General options:\n"
        m+=dict_to_str(self.option_d)
        return m

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~PRIVATE METHODS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    def _split_reads (self, in_q, error_q):
        """
        Mono-threaded reader
        """
        self.log.debug("\t[split_reads] Start reading input file/stream")
        try:
            # Open input file or stdin if 0
            with open (self.input_fn) as fp:

                # Get header line and extract corresponding index
                input_header = fp.readline().rstrip().split("\t")
                if not input_header:
                    raise NanopolishCompError ("Input file/stream is empty")

                idx = self._get_field_idx (input_header)
                n_reads = 0

                # First data line exception
                read_l = []
                event_l = fp.readline().rstrip().split("\t")
                cur_read_id = event_l[idx["read_id"]]
                cur_ref_id = event_l[idx["ref_id"]]
                event_d = self._event_list_to_dict (event_l, idx)
                read_l.append (event_d)

                for line in fp:
                    # Early ending if required
                    if self.max_reads and n_reads == self.max_reads:
                        break
                    # Get event line
                    event_l = line.rstrip().split("\t")
                    read_id = event_l[idx["read_id"]]
                    ref_id = event_l[idx["ref_id"]]
                    event_d = self._event_list_to_dict (event_l, idx)

                    # Line correspond to the same ids
                    if read_id != cur_read_id or ref_id != cur_ref_id:
                        in_q.put ((cur_read_id, cur_ref_id, read_l))
                        n_reads+=1
                        read_l = []
                        cur_read_id = read_id
                        cur_ref_id = ref_id

                    # In any case extend list corresponding to current read_id/ref_id
                    read_l.append(event_d)

                # Last data line exception
                in_q.put ((cur_read_id, cur_ref_id, read_l))
                n_reads+=1

        # Manage exceptions and deal poison pills
        except Exception:
            error_q.put (NanopolishCompError(traceback.format_exc()))
        finally:
            for i in range (self.threads):
                in_q.put(None)
            self.log.debug("\t[split_reads] Done")

    def _process_read (self, in_q, out_q, error_q, pid):
        """
        Multi-threaded workers
        """
        self.log.debug("\t[process_read {}] Starting processing reads".format(pid))
        try:
            # Collapse event at kmer level
            for read_id, ref_id, read_l in iter(in_q.get, None):

                # Write read header to str
                read_str = "#{}\t{}\n".format(read_id, ref_id)
                read_str+= "{}\n".format (self._make_ouput_header(event_d=read_l[0]))

                # Init values for first kmer
                kmer_d = self._init_kmer_dict(event_d=read_l[0])

                # Init read dictionary
                read_d = OrderedDict ()
                read_d["read_id"] = read_id
                read_d["ref_id"] = ref_id
                read_d["dwell_time"] = 0.0
                read_d["kmers"] = 0
                read_d["NNNNN_kmers"] = 0
                read_d["mismatch_kmers"] = 0
                read_d["missing_kmers"] = 0
                read_d["ref_start"] = kmer_d["ref_pos"]

                # Iterate over the rest of the lines
                for event_d in read_l [1:]:
                    pos_offset = event_d["ref_pos"]-kmer_d["ref_pos"]

                    # Same position = update current kmer
                    if pos_offset == 0:
                        kmer_d = self._update_kmer_dict(kmer_d=kmer_d, event_d=event_d)

                    # New position = write previous kmer and start new one
                    else:
                        # Update read counter
                        read_d["dwell_time"] += kmer_d["dwell_time"]
                        if kmer_d["NNNNN_dwell_time"]:
                            read_d["NNNNN_kmers"] += 1
                        if kmer_d ["mismatch_dwell_time"]:
                            read_d["mismatch_kmers"] += 1
                        if pos_offset >=2:
                            read_d["missing_kmers"] += (pos_offset-1)
                        read_d["kmers"] += 1
                        # Converts previous kmer to str and init new kmer
                        read_str += "{}\n".format(self._kmer_dict_to_str(kmer_d=kmer_d))
                        kmer_d = self._init_kmer_dict(event_d=event_d)

                # Last read_d update
                read_d["dwell_time"] += kmer_d["dwell_time"]
                if kmer_d ["NNNNN_dwell_time"]:
                    read_d["NNNNN_kmers"] += 1
                if kmer_d ["mismatch_dwell_time"]:
                    read_d["mismatch_kmers"] += 1
                if pos_offset >=2:
                    read_d["missing_kmers"] += (pos_offset-1)
                read_d["ref_end"] = kmer_d["ref_pos"]+1
                read_d["kmers"] += 1

                # Last kmer
                read_str += "{}\n".format(self._kmer_dict_to_str(kmer_d=kmer_d))

                # Add the current read details to queue
                out_q.put((read_d, read_str))

        # Manage exceptions and deal poison pills
        except Exception:
            error_q.put (NanopolishCompError(traceback.format_exc()))
        finally:
            self.log.debug("\t[process_read {}] Done".format(pid))
            out_q.put(None)

    def _write_output (self, out_q, error_q):
        """
        Mono-threaded Writer
        """
        self.log.debug("\t[write_output] Start rwriting output")

        byte_offset = n_reads = 0
        t = time()

        try:
            # Open output files
            data_fn = os.path.join(self.outdir, self.outprefix+"_eventalign_collapse.tsv")
            idx_fn = os.path.join(self.outdir, self.outprefix+"_eventalign_collapse.tsv.idx")
            with open (data_fn, "w") as data_fp,\
                 open (idx_fn, "w") as idx_fp,\
                 tqdm (unit=" reads", mininterval=0.1, smoothing=0.1, disable=self.log.level>=30) as pbar:

                idx_fp.write ("ref_id\tref_start\tref_end\tread_id\tkmers\tdwell_time\tNNNNN_kmers\tmismatch_kmers\tmissing_kmers\tbyte_offset\tbyte_len\n")

                n_reads = 0
                for _ in range (self.threads):
                    for (read_d, read_str) in iter (out_q.get, None):
                        byte_len = len(read_str)

                        data_fp.write (read_str)
                        idx_fp.write ("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format (
                            read_d["ref_id"],
                            read_d["ref_start"],
                            read_d["ref_end"],
                            read_d["read_id"],
                            read_d["kmers"],
                            read_d["dwell_time"],
                            read_d["NNNNN_kmers"],
                            read_d["mismatch_kmers"],
                            read_d["missing_kmers"],
                            byte_offset,
                            byte_len-1))

                        byte_offset += byte_len
                        n_reads += 1
                        if self.log.level<30:
                            pbar.update(1)

                # Flag last line
                data_fp.write ("#\n")

            # Open log file
            log_fn = os.path.join(self.outdir, self.outprefix+"_eventalign_collapse.log")
            with open (log_fn, "w") as log_fp:
                log_fp.write (str(self))

        # Manage exceptions and deal poison pills
        except Exception:
            error_q.put (NanopolishCompError(traceback.format_exc()))
        finally:
            self.log.debug("\t[write_output] Done")
            self.log.warning ("[Eventalign_collapse] total reads: {} [{} reads/s]\n".format(n_reads, round (n_reads/(time()-t), 2)))
            error_q.put(None)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~HELPER PRIVATE METHODS~~~~~~~~~~~~~~~~~~~~~~~~~~#

    def _get_field_idx (self, input_header):
        """"""
        # Get index of fields to fetch
        idx = OrderedDict()
        idx["ref_id"] = input_header.index ("contig")
        if "read_name" in input_header:
            idx["read_id"] = input_header.index ("read_name")
        elif "read_index" in input_header:
            idx["read_id"] = input_header.index ("read_index")
        idx["ref_pos"] = input_header.index ("position")
        idx["ref_kmer"] = input_header.index ("reference_kmer")
        idx["mod_kmer"] = input_header.index ("model_kmer")
        idx["event_len"] = input_header.index ("event_length")
        # Facultative field start and end index
        if "start_idx" in input_header and "end_idx" in input_header:
            idx["start_idx"] = input_header.index ("start_idx")
            idx["end_idx"] = input_header.index ("end_idx")
        # Facultative field samples
        if "samples" in input_header:
            idx["samples"] = input_header.index ("samples")
        return idx

    def _event_list_to_dict (self, event_l, idx):
        """Get interesting fields from event list and cast in appropriate type"""
        event_d = OrderedDict()
        event_d["ref_pos"] = int(event_l[idx["ref_pos"]])
        event_d["ref_kmer"] = event_l[idx["ref_kmer"]]
        event_d["mod_kmer"] = event_l[idx["mod_kmer"]]
        event_d["event_len"] = float(event_l[idx["event_len"]])
        if "start_idx" in idx:
            event_d["start_idx"] = int(event_l[idx["start_idx"]])
            event_d["end_idx"] = int(event_l[idx["end_idx"]])
        if "samples" in idx:
            event_d["sample_list"] = event_l[idx["samples"]].split(",")
        return event_d

    def _init_kmer_dict (self, event_d):
        """Start a new kmer dict from first event values"""
        kmer_d = OrderedDict ()
        kmer_d["ref_pos"] = event_d["ref_pos"]
        kmer_d["ref_kmer"] = event_d["ref_kmer"]
        kmer_d["num_events"] = 1
        kmer_d["dwell_time"] = event_d["event_len"]
        kmer_d["NNNNN_dwell_time"] = 0.0
        kmer_d["mismatch_dwell_time"] = 0.0
        if event_d["mod_kmer"] == "NNNNN":
            kmer_d["NNNNN_dwell_time"] += event_d["event_len"]
        elif event_d["mod_kmer"] != event_d["ref_kmer"]:
            kmer_d["mismatch_dwell_time"] += event_d["event_len"]
        if "start_idx" in event_d:
            kmer_d["start_idx"] = event_d["start_idx"]
            kmer_d["end_idx"] = event_d["end_idx"]
        if "sample_list" in event_d:
            kmer_d["sample_list"] = event_d["sample_list"]
        return kmer_d

    def _update_kmer_dict (self, kmer_d, event_d):
        """Update kmer dict from subsequent event values"""
        kmer_d["num_events"] += 1
        kmer_d["dwell_time"] += event_d["event_len"]
        if event_d["mod_kmer"] == "NNNNN":
            kmer_d["NNNNN_dwell_time"] += event_d["event_len"]
        elif event_d["mod_kmer"] != event_d["ref_kmer"]:
            kmer_d["mismatch_dwell_time"] += event_d["event_len"]
        if "start_idx" in event_d:
            kmer_d["start_idx"] = event_d["start_idx"]
        if "sample_list" in event_d:
            kmer_d["sample_list"].extend(event_d["sample_list"])
        return kmer_d

    def _kmer_dict_to_str (self, kmer_d):
        """"""
        # Write base fields
        s = "{}\t{}\t{}\t{}\t{}\t{}".format(
            kmer_d["ref_pos"],
            kmer_d["ref_kmer"],
            kmer_d["num_events"],
            kmer_d["dwell_time"],
            kmer_d["NNNNN_dwell_time"],
            kmer_d["mismatch_dwell_time"])
        # Facultative index fields
        if "start_idx" in kmer_d:
            s += "\t{}\t{}".format(
                kmer_d["start_idx"],
                kmer_d["end_idx"])
        # Facultative samples fields
        if "sample_list" in kmer_d:
            sample_array = np.array (kmer_d["sample_list"], dtype=np.float32)
            if "mean" in self.stat_fields:
                s += "\t{}".format(np.mean (sample_array))
            if "std" in self.stat_fields:
                s += "\t{}".format(np.std (sample_array))
            if "median" in self.stat_fields:
                s += "\t{}".format(np.median (sample_array))
            if "mad" in self.stat_fields:
                s += "\t{}".format(np.median(np.abs(sample_array-np.median(sample_array))))
            if "num_signals" in self.stat_fields:
                s += "\t{}".format(len(sample_array))
            if self.write_samples:
                s += "\t{}".format(",".join(kmer_d["sample_list"]))
        return s

    def _make_ouput_header (self, event_d):
        """"""
        # Write base fields
        s = "ref_pos\tref_kmer\tnum_events\tdwell_time\tNNNNN_dwell_time\tmismatch_dwell_time"
        # Write extra fields
        if "start_idx" in event_d:
            s += "\tstart_idx\tend_idx"
        if "sample_list" in event_d:
            if "mean" in self.stat_fields:
                s += "\tmean"
            if "std" in self.stat_fields:
                s += "\tstd"
            if "median" in self.stat_fields:
                s += "\tmedian"
            if "mad" in self.stat_fields:
                s += "\tmad"
            if "num_signals" in self.stat_fields:
                s += "\tnum_signals"
            if self.write_samples:
                s += "\tsamples"
        return s
